-- This file is part of SmartEiffel The GNU Eiffel Compiler Tools and Libraries
--
-- SmartEiffel is  free software;  you can redistribute it and/or  modify it
-- under  the terms of the  GNU General Public License, as published by  the
-- Free Software Foundation; either version 2, or (at your option) any later
-- version.
-- SmartEiffel is distributed in the hope that it will be useful but WITHOUT 
-- ANY WARRANTY;  without  even the implied warranty  of MERCHANTABILITY  or
-- FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
-- more details.  You should have received a copy of  the GNU General Public
-- License along with SmartEiffel;  see the file COPYING.  If not,  write to
-- the Free Software Foundation,  Inc., 59 Temple Place - Suite 330,  Boston, 
-- MA 02111-1307, USA.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.
--                           - University of Nancy 1 - FRANCE
-- Copyright(C) 2003:      INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne
--                           - University of Nancy 2 - FRANCE
--
--                 Dominique COLNET, Suzanne COLLIN, Olivier ZENDRA,
--                           Philippe RIBET, Cyril ADRIAN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
--
class REAL_CONSTANT
   --
   -- For Manifest Constant REAL as well as DOUBLE.
   --

inherit BASE_TYPE_CONSTANT

creation make, from_double

feature

   is_static: BOOLEAN is False

   to_string: STRING
         -- Cleanly written view of the constant (ANSI C compatible).

   is_infinity: BOOLEAN is
      do
         Result := to_string.has_substring (once "inf") or else
                   to_string.has_substring (once "INF")
         -- ISO C99 compilers generate "inf" or "infinity", optionally preceded
         -- by a minus sign. Visual C++ 7.1 generates something else, but
         -- this should catch it, see ticket:42
      end

   to_integer_or_error: INTEGER is
      do
         to_integer_error
      end

   value: DOUBLE is
      do
         Result := to_string.to_double
      end

   compile_to_c is
      do
         if not is_infinity then
             cpp.put_string(to_string)
         elseif to_string.first = '-' then
             cpp.put_string(once "(/*minus infinity */-1.0/0.0)")
         else
             cpp.put_string(once "(/*plus infinity */+1.0/0.0)")
         end
         -- There is no standard way to write a floating point
         -- infinity in C. ISO C99 introduces an INFINITY macro
         -- but we should not depend on having an ISO C99 compiler.
         -- Perhaps this can be solved with an "#ifdef" at the runtime.
      end

   compile_target_to_jvm, compile_to_jvm is
      do
         if result_type.is_double then
            code_attribute.opcode_push_as_double(to_string)
         else
            code_attribute.opcode_push_as_float(to_string)
         end
      end
         
   jvm_branch_if_false: INTEGER is
      do
      end

   jvm_branch_if_true: INTEGER is
      do
      end

   compile_to_jvm_into(dest: E_TYPE): INTEGER is
      do
         if dest.is_real then
            code_attribute.opcode_push_as_float(to_string)
            Result := 1
         elseif dest.is_double then
            code_attribute.opcode_push_as_double(to_string)
            Result := 2
         else
            Result := standard_compile_to_jvm_into(dest)
         end
      end

   to_runnable(ct: E_TYPE): like Current is
      do
         Result := Current
      end

   simplify_1_, simplify_2: like Current is
      do
         Result := Current
      end
   
feature {IMPLICIT_CAST, TMP_FEATURE}

   set_result_type(rt: E_TYPE) is
      require
         rt.is_double or else rt.is_real
      do
         result_type := rt
      ensure
         result_type = rt
      end

feature {REAL_CONSTANT_VISITOR}

   accept(visitor: REAL_CONSTANT_VISITOR) is
      do
         visitor.visit_real_constant(Current)
      end

feature {NONE}

   make (sp: like start_position; ts: like to_string) is
      require
         not sp.is_unknown
         ts.is_double
         string_aliaser.item(ts) /= ts
      local
         pos, exponent: INTEGER
         stop: BOOLEAN
      do
         start_position := sp
         to_string := ts
         if smart_eiffel.pretty_flag then
         elseif smart_eiffel.short_flag then
         elseif is_infinity then
         else
            -- Kind of normalisation to allow easy comparison.
            -- *** This seems to be used in CODE_ATTRIBUTE and not 
            -- *** much more. Maybe we should get rid of the whole 
            -- *** normalisation. <FM-14/08/2004>
            -- Part 1: Remove the exponent if there is one.
            pos := to_string.first_index_of('e')
            if pos = 0 then
               pos := to_string.first_index_of('E')
            end
            if pos /= 0 then
               exponent := to_string.substring(pos + 1, to_string.count).to_integer
               to_string.keep_head(pos-1)
            end
            -- Part 2: find out if there is a decimal point. Remove it
            -- if we'll need to move it because of the exponent.
            pos := to_string.first_index_of('.')
            if exponent /= 0 and then pos /= 0 then
               exponent := exponent - (to_string.count - pos)
               to_string.remove(pos)
               pos := 0
            end
            -- Part 3: normalize the decimal part of the number.
            if pos /= 0 then
               -- If there is a decimal point, just get rid of 
               -- trailing zeroes.
               check
                  exponent = 0
               end
               from
               until
                  to_string.last /= '0'
               loop
                  to_string.remove_last(1)
               end
               if to_string.last = '.' then
                  to_string.extend('0')
               end
            else
               -- If there is no decimal point, add one. Decide what 
               -- to do with the trailing zeroes according to the exponent.
               if exponent >= 0 then
                  to_string.extend_multiple('0', exponent)
                  to_string.append(once ".0")
               else
                  from
                     stop := False
                  until
                     stop
                  loop
                     if to_string.last = '0' then
                        if to_string.count > 1 then
                           to_string.remove_last(1)
                           exponent := exponent + 1
                        else
                           exponent := 0
                        end
                        if exponent = 0 then
                           to_string.append(once ".0")
                           stop := True
                        end
                     else
                        pos := to_string.count + exponent + 1
                        if pos <= 1 then
                           to_string.precede_multiple('0', 3 - pos)
                           pos := 2
                           to_string.put('.', pos)
                        else
                           to_string.insert_character('.', pos)
                        end
                        stop := True
                     end
                  end
               end
            end
            -- Part 4: normalize the integral part of the number.
            if pos = 1 then
               check
                  to_string.first_index_of('.') = 1
               end
               to_string.precede('0')
            else
               check
                  to_string.first_index_of('.') /= 1
               end
               from
                  pos := 1
               until
                  to_string.item(pos) /= '0'
               loop
                  pos := pos + 1
               end
               if to_string.item(pos) = '.' then
                  pos := pos - 1
               end
               if pos > 1 then
                  to_string.remove_head(pos - 1)
               end
            end
         end
         -- Setting the `result_type':
         create {TYPE_REAL} result_type.make(start_position, as_real)
      ensure
         start_position = sp
         to_string = ts
      end

   from_double(sp: like start_position; double: DOUBLE; rt: like result_type) is
      require
         rt.is_double or else rt.is_real
      do
         make(sp, double.to_string_format(10))
         --                               **
         -- We need to redesign REAL/DOUBLE notation... :-(
         result_type := rt
      end
   
end -- REAL_CONSTANT
